# -*- coding: utf-8 -*-

import datetime

from construct import (Adapter, Array, BitsInteger, BitStruct, Bitwise, Bytes,
                       ByteSwapped, Bytewise, Computed, Construct, Container,
                       Default, ExprSymmetricAdapter, Flag, Int24ub,
                       ListContainer, Struct, Subconstruct, SymmetricAdapter,
                       obj_, this)


class DateAdapter(Adapter):
    def _decode(self, obj, context, path):
        return datetime.datetime(
            obj[0] * 100 + obj[1],
            obj[2],
            obj[3],
            obj[4],
            obj[5],
            obj[6] if len(obj) > 6 else 0,
        )

    def _encode(self, obj, context, path):
        bts = [
            obj.year // 100,
            obj.year % 100,
            obj.month,
            obj.day,
            obj.hour,
            obj.minute,
        ]

        if self.subcon.length == 7:
            bts.append(obj.second)

        return bytes(bts)


class DictArray(Adapter):
    def __init__(self, count, first_index, subcon, pick_key: str = None):
        super(DictArray, self).__init__(Array(count, subcon))
        self._first_index = first_index
        self._pick_key = pick_key

    def _decode(self, obj, context, path):
        if not isinstance(obj, list):
            raise TypeError("list should be passed for decoding")

        n_obj = Container()
        for item in obj:
            v = item
            if self._pick_key:
                v = item.get(self._pick_key)
            n_obj[item._index] = v

        return n_obj

    def _encode(self, obj, context, path):
        if not isinstance(obj, dict):
            raise TypeError("dict should be passed for decoding")

        n_obj = ListContainer()

        count = self.subcon.count  # Array count

        for k in range(self._first_index, self._first_index + count):
            o = {"_index": k}
            v = obj.get(k)
            if v is not None:
                if self._pick_key:
                    v = {self._pick_key: v}
                o.update(v)
            n_obj.append(o)

        return n_obj


class EnumerationAdapter(Subconstruct):
    def __init__(self, subcon):
        super(EnumerationAdapter, self).__init__(subcon)

        def find_count(s):
            if hasattr(s, "count"):
                return s.count
            else:
                return find_count(s.subcon)

        self.size = find_count(subcon)

    def _build(self, obj, stream, context, path):
        zones = list([i in obj for i in range(1, self.size + 1)])

        return self.subcon._build(zones, stream, context, path)


def StatusFlags(count, start_index_from=1):
    return DictArray(
        count,
        start_index_from,
        Struct(
            "_index" / Computed(this._index + start_index_from),
            "flag" / Default(Flag, False),
        ),
        pick_key="flag",
    )


def ModuleTroubles(count, start_index_from=1):
    return DictArray(
        count=count,
        first_index=start_index_from,
        subcon=BitStruct(
            "_index" / Computed(this._index + start_index_from),
            "aux_trouble" / Flag,
            "battery_fail" / Flag,
            "ac_trouble" / Flag,
            "printer_trouble" / Flag,
            "fail_to_comm" / Flag,
            "tlm_trouble" / Flag,
            "rom_error" / Flag,
            "tamper" / Flag,
        ),
    )


def ZoneFlags(count, start_index_from=1):
    return DictArray(
        count,
        start_index_from,
        BitStruct(
            "_index" / Computed(this._index + start_index_from),
            "generated_alarm" / Default(Flag, False),
            "presently_in_alarm" / Default(Flag, False),
            "activated_entry_delay" / Default(Flag, False),
            "activated_intellizone_delay" / Default(Flag, False),
            "bypassed" / Default(Flag, False),
            "shutted_down" / Default(Flag, False),
            "tx_delay" / Default(Flag, False),
            "supervision_trouble" / Default(Flag, False),
        ),
    )


def PGMFlags(count, start_index_from=1):
    return DictArray(
        count,
        start_index_from,
        Bitwise(
            Struct(
                "_index" / Computed(this._index + start_index_from),
                "fire_2_wires" / Default(Flag, False),
                "normally_closed" / Default(Flag, False),
                "_unknown1" / BitsInteger(1),
                "disabled"
                / ExprSymmetricAdapter(
                    Default(Flag, False), lambda obj, ctx: not obj
                ),  #  False when a relay is assigned
                "_unknown2" / BitsInteger(2),
                "timer_active" / Default(Flag, False),  # when timer is active
                "on" / Default(Flag, False),  # when is activated
                "time_left"
                / Bytewise(ByteSwapped(Default(Int24ub, 0))),  # byte in seconds
            ),
        ),
    )


ZoneFlagBitStruct = BitStruct(
    "generated_alarm" / Default(Flag, False),
    "presently_in_alarm" / Default(Flag, False),
    "activated_entry_delay" / Default(Flag, False),
    "activated_intellizone_delay" / Default(Flag, False),
    "bypassed" / Default(Flag, False),
    "shutted_down" / Default(Flag, False),
    "tx_delay" / Default(Flag, False),
    "supervision_trouble" / Default(Flag, False),
)


class PartitionStatus(Subconstruct):
    first2 = BitStruct(
        "fire_alarm" / Flag,
        "audible_alarm" / Flag,
        "silent_alarm" / Flag,
        "was_in_alarm" / Flag,  # is in alarm
        "arm_no_entry" / Flag,
        "arm_stay" / Flag,  # Armed in Stay mode
        "arm_away" / Flag,  # Armed in Away mode
        "arm" / Flag,  # Armed
        "lockout" / Flag,
        "programming" / Flag,
        "zone_bypassed" / Flag,
        "alarm_in_memory" / Flag,
        "trouble" / Flag,
        "entry_delay" / Flag,
        "exit_delay" / Flag,
        "ready" / Flag,
    )

    last4 = BitStruct(
        "zone_supervision_trouble" / Flag,
        "zone_fire_loop_trouble" / Flag,
        "zone_low_battery_trouble" / Flag,
        "zone_tamper_trouble" / Flag,
        "voice_arming" / Flag,
        "auto_arming_engaged" / Flag,
        "fire_delay_in_progress" / Flag,
        "intellizone_engage" / Flag,
        "time_to_refresh_zone_status" / Flag,
        "panic_alarm" / Flag,
        "police_code_delay" / Flag,  # Within police code delay
        "follow_become_delay" / Flag,  # Follow become delay when is bypassed
        "remote_arming" / Flag,
        "stay_arming_auto" / Flag,  # if no entry zone is tripped
        "partition_recently_close" / Flag,
        "cancel_alarm_reporting_on_disarming" / Flag,
        "tx_delay_finished" / Flag,  # (Time Out / instant alarm)
        "auto_arm_reach" / Flag,
        "fire_delay_end" / Flag,
        "no_movement_delay_end" / Flag,
        "alarm_duration_finished" / Flag,
        "entry_delay_finished" / Flag,
        "exit_delay_finished" / Flag,
        "intellizone_delay_finished" / Flag,
        "_free0" / BitsInteger(3),
        "all_zone_closed" / Flag,  # (Bypass or not)
        "inhibit_ready" / Flag,
        "bypass_ready" / Flag,
        "force_ready" / Flag,
        "stay_instant_ready" / Flag,
    )

    def __init__(self, subcons_or_size):
        if isinstance(subcons_or_size, Construct):
            self.size = subcons_or_size.sizeof()
            subcons = subcons_or_size
        else:
            self.size = subcons_or_size
            subcons = Bytes(subcons_or_size)
        super(PartitionStatus, self).__init__(subcons)

    def _parse(self, stream, context, path):
        obj = Container()

        if self.size == 32:
            for i in range(1, 6):
                obj[i] = self.first2._parsereport(stream, context, path)
                obj[i].update(self.last4._parsereport(stream, context, path))

            obj[6] = self.first2._parsereport(stream, context, path)
        elif self.size == 16:
            obj[6] = self.last4._parsereport(stream, context, path)
            for i in range(7, 9):
                obj[i] = self.first2._parsereport(stream, context, path)
                obj[i].update(self.last4._parsereport(stream, context, path))
        else:
            raise Exception("Not supported size. Only 32 or 16")

        return obj


class EventAdapter(Adapter):
    def _decode(self, obj, context, path):
        event_group = obj[0]
        event_1_high_nibble = obj[1] >> 6
        event_2_high_nibble = (obj[1] >> 4) & 0b11
        event_1 = obj[2] + (event_1_high_nibble << 8)
        event_2 = obj[3] + (event_2_high_nibble << 8)
        partition = obj[1] & 0x0F

        return Container(
            {
                "major": event_group,
                "minor": event_1,
                "minor2": event_2,
                "partition": partition,
            }
        )

    def _encode(self, obj, context, path):
        return bytes(
            [
                obj.major,
                obj.partition & 0x0F
                | (obj.minor >> 8 & 3) << 6
                | (obj.minor2 >> 8 & 3) << 4,
                obj.minor & 0xFF,
                obj.minor2 & 0xFF,
            ]
        )


# class CompressedEventAdapter(Adapter):
#     def _decode(self, obj, context, path):
#         day = obj[0] >> 3
#         month = (obj[0] & 0b111) << 1 | obj[1] >> 7
#         century = obj[1] & 0b1111111
#         year = century * 100 + (obj[2] >> 1)
#         hour = (obj[2] & 0b1) << 4 | obj[3] >> 4
#         minute = (obj[3] & 0b1111) << 2 | obj[4] >> 6
#
#         Container({
#             'time': datetime.datetime(year, month, day, hour, minute)
#         })
#
#     def _encode(self, obj, context, path):
#         return [obj.year / 100, obj.year % 100, obj.month, obj.day, obj.hour, obj.minute, obj.second]



# coding: utf-8

import sys

from setuptools import find_packages, setup

from paradox import VERSION

needs_pytest = {"pytest", "test", "ptr"}.intersection(sys.argv)
pytest_runner = ["pytest-runner"] if needs_pytest else []

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="paradox-alarm-interface",
    version=VERSION,
    author="João Paulo Barraca",
    author_email="jpbarraca@gmail.com",
    description="Interface to Paradox Alarm Panels",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ParadoxAlarmInterface/pai",
    download_url=f"https://github.com/ParadoxAlarmInterface/pai/archive/{VERSION}.tar.gz",
    packages=find_packages(exclude=["tests", "tests.*", "config.*", "docs.*"]),
    install_requires=["construct~=2.9.43", "argparse>=1.4.0"],
    python_requires=">=3.6",
    setup_requires=["wheel"] + pytest_runner,
    tests_require=[
        "pytest",
        "pytest-asyncio",
        "pytest-env",
        "pytest-mock",
        "mock",
        "asynctest",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Eclipse Public License 2.0 (EPL-2.0)",
        "Operating System :: OS Independent",
    ],
    extras_require={
        "YAML": ["pyyaml>=5.2.0"],
        "Serial": ["pyserial-asyncio>=0.4"],
        "IP": ["requests>=2.20.0"],
        "MQTT": ["paho_mqtt>=1.5.0"],
        "Pushbullet": ["pushbullet.py>=0.11.0", "ws4py>=0.4.2"],
        "Pushover": ["chump>=1.6.0"],
        "Signal": ["pygobject>=3.20.0", "pydbus>=0.6.0", "gi>=1.2"],
    },
    entry_points={
        "console_scripts": [
            "ip150-connection-decrypt = paradox.console_scripts.ip150_connection_decrypt:main [YAML]",
            "pai-service = paradox.console_scripts.pai_run:main",
            "pai-dump-memory = paradox.console_scripts.pai_dump_memory:main",
        ]
    },
    license="EPL",
)



import logging
from enum import IntEnum


class RunState(IntEnum):
    STOP = 0
    INIT = 1
    CONNECTED = 2
    RUN = 3
    PAUSE = 4
    ERROR = 5


class EventLevel(IntEnum):
    NOTSET = logging.NOTSET
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARN = logging.WARN
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL

    @staticmethod
    def from_name(level="NOTSET"):
        if isinstance(level, EventLevel):
            return level

        level = level.upper()
        if level in EventLevel.__members__:
            return EventLevel.__members__[level]

        raise (
            Exception(
                "Invalid log level {}. Valid levels: {}".format(
                    level, list(EventLevel.__members__)
                )
            )
        )

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value

        if other.__class__ is int:
            return self.value < other

        return NotImplemented

    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return self.value > other.value

        if other.__class__ is int:
            return self.value > other

        return NotImplemented

    def __le__(self, other):
        if self.__class__ is other.__class__:
            return self.value <= other.value

        if other.__class__ is int:
            return self.value <= other

        return NotImplemented

    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.value >= other.value

        if other.__class__ is int:
            return self.value >= other

        return NotImplemented

    def __str__(self):
        return self.name



# -*- coding: utf-8 -*-

import logging
import re

import chump

from paradox.config import config as cfg
from paradox.event import EventLevel
from paradox.interfaces.text.core import ConfiguredAbstractTextInterface

logger = logging.getLogger("PAI").getChild(__name__)

_level_2_priority = {
    EventLevel.NOTSET: chump.LOWEST,
    EventLevel.DEBUG: chump.LOWEST,
    EventLevel.INFO: chump.LOW,
    EventLevel.WARN: chump.NORMAL,
    EventLevel.ERROR: chump.HIGH,
    EventLevel.CRITICAL: chump.EMERGENCY,
}


class PushoverTextInterface(ConfiguredAbstractTextInterface):
    """Interface Class using Pushover"""

    def __init__(self, alarm):
        super().__init__(
            alarm,
            cfg.PUSHOVER_EVENT_FILTERS,
            cfg.PUSHOVER_ALLOW_EVENTS,
            cfg.PUSHOVER_IGNORE_EVENTS,
            cfg.PUSHOVER_MIN_EVENT_LEVEL,
        )

        self.app = None
        self.users = {}

    def _run(self):
        super(PushoverTextInterface, self)._run()

        self.app = chump.Application(cfg.PUSHOVER_KEY)
        if not self.app.is_authenticated:
            raise Exception(
                "Failed to authenticate with Pushover. Please check PUSHOVER_APPLICATION_KEY"
            )

    def send_message(self, message: str, level: EventLevel):
        for user_key, devices_raw in cfg.PUSHOVER_BROADCAST_KEYS.items():
            user = self.users.get(user_key)  # type: chump.User

            if user is None:
                user = self.users[user_key] = self.app.get_user(user_key)

            if not user.is_authenticated:
                raise Exception(
                    "Failed to check user key with Pushover. Please check PUSHOVER_BROADCAST_KEYS[%s]"
                    % user_key
                )

            if devices_raw == "*" or devices_raw is None:
                try:
                    user.send_message(
                        message,
                        title="Alarm",
                        priority=_level_2_priority.get(level, chump.NORMAL),
                    )
                except:
                    logger.exception("pushbullet send message")

            else:
                devices = list(filter(bool, re.split(r"[\s]*,[\s]*", devices_raw)))

                for elem in (elem for elem in devices if elem not in user.devices):
                    logger.warning(
                        "%s is not in the Pushover device list for the user %s"
                        % (elem, user_key)
                    )

                for device in devices:
                    try:
                        user.send_message(
                            message,
                            title="PAI",
                            device=device,
                            priority=_level_2_priority.get(level, chump.NORMAL),
                        )
                    except:
                        logger.exception("pushbullet send message")

        # TODO: Missing the message reception
